# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch

from typing import ClassVar, Tuple, List
from decimal import Decimal
from typing import Optional

from .step import ViewsUpdate, MetricsUpdate, Step, StepError, DeferredStepError, State

from ..logging import err, warn, info, debug, verbose
from ..config import Variable


class MetricChecker(Step):
    """
    Raises a (deferred) error if a Decimal metric exceeds a certain threshold.
    """

    inputs = []
    outputs = []

    metric_name: ClassVar[str] = NotImplemented
    metric_description: ClassVar[str] = NotImplemented
    deferred: ClassVar[bool] = True

    @classmethod
    def get_help_md(Self, **kwargs):  # pragma: no cover
        threshold_string = Self.get_threshold_description(None)
        if threshold_string is None:
            threshold_string = str(Self.get_threshold(None))
        dynamic_docstring = "Raises"
        if Self.deferred:
            dynamic_docstring += " a deferred error"
        else:
            dynamic_docstring += " an immediate error"
        dynamic_docstring += f" if {Self.metric_description} (metric: ``{Self.metric_name}``) are >= {threshold_string}."

        return super().get_help_md(docstring_override=dynamic_docstring, **kwargs)

    def get_threshold(self: Optional["MetricChecker"]) -> Optional[Decimal]:
        return Decimal(0)

    def get_threshold_description(self: Optional["MetricChecker"]) -> Optional[str]:
        return None

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        threshold = self.get_threshold()

        if threshold is None:
            warn(
                f"Threshold for {self.metric_description} is not set. The checker will be skipped."
            )
        else:
            metric_value = state_in.metrics.get(self.metric_name)
            if metric_value is not None:
                if metric_value > threshold:
                    error_msg = f"{metric_value} {self.metric_description} found."
                    if self.deferred:
                        err(f"{error_msg} - deferred")
                        raise DeferredStepError(error_msg)
                    else:
                        err(f"{error_msg}")
                        raise StepError(error_msg)

                else:
                    info(f"Check for {self.metric_description} clear.")
            else:
                warn(
                    f"The {self.metric_description} metric was not found. Are you sure the relevant step was run?"
                )

        return {}, {}


@Step.factory.register()
class YosysUnmappedCells(MetricChecker):
    id = "Checker.YosysUnmappedCells"
    name = "Unmapped Cells Checker"
    deferred = False

    metric_name = "design__instance_unmapped__count"
    metric_description = "Unmapped Yosys instances"


@Step.factory.register()
class YosysSynthChecks(MetricChecker):
    id = "Checker.YosysSynthChecks"
    name = "Yosys Synth Checks"
    deferred = False

    metric_name = "synthesis__check_error__count"
    metric_description = "Yosys check errors"


@Step.factory.register()
class TrDRC(MetricChecker):
    id = "Checker.TrDRC"
    name = "Routing DRC Checker"
    long_name = "Routing Design Rule Checker"

    metric_name = "route__drc_errors"
    metric_description = "Routing DRC errors"


@Step.factory.register()
class MagicDRC(MetricChecker):
    id = "Checker.MagicDRC"
    name = "Magic DRC Checker"
    long_name = "Magic Design Rule Checker"

    metric_name = "magic__drc_error__count"
    metric_description = "Magic DRC errors"


@Step.factory.register()
class IllegalOverlap(MetricChecker):
    id = "Checker.IllegalOverlap"
    name = "Illegal Overlap Checker"
    long_name = "Spice Extraction-based Illegal Overlap Checker"

    metric_name = "magic__illegal_overlap__count"
    metric_description = "Magic Illegal Overlap errors"


@Step.factory.register()
class DisconnectedPins(MetricChecker):
    id = "Checker.DisconnectedPins"
    name = "Disconnected Pins Checker"
    deferred = False

    metric_name = "design__critical_disconnected_pin__count"
    metric_description = "critical disconnected pins"


@Step.factory.register()
class WireLength(MetricChecker):
    id = "Checker.WireLength"
    name = "Wire Length Threshold Checker"

    metric_name = "route__wirelength__max"
    metric_description = "Threshold-surpassing long wires"

    def get_threshold(self) -> Optional[Decimal]:
        threshold = self.config["WIRE_LENGTH_THRESHOLD"]
        assert threshold is None or isinstance(threshold, Decimal)
        return threshold

    def get_threshold_description(self) -> Optional[str]:
        return "the threshold specified in the configuration file."


@Step.factory.register()
class XOR(MetricChecker):
    id = "Checker.XOR"
    name = "XOR Difference Checker"
    long_name = "Magic vs. KLayout XOR Difference Checker"

    metric_name = "design__xor_difference__count"
    metric_description = "XOR differences"


@Step.factory.register()
class LVS(MetricChecker):
    id = "Checker.LVS"
    name = "LVS Error Checker"
    long_name = "Layout vs. Schematic Error Checker"

    metric_name = "design__lvs_error__count"
    metric_description = "LVS errors"


@Step.factory.register()
class PowerGridViolations(MetricChecker):
    id = "Checker.PowerGridViolations"
    name = "Power Grid Violation Checker"

    metric_name = "design__power_grid_violation__count"
    metric_description = "power grid violations (as reported by OpenROAD PSM- you may ignore these if LVS passes)"


@Step.factory.register()
class LintErrors(MetricChecker):
    id = "Checker.LintErrors"
    name = "Lint Errors Checker"
    long_name = "Lint Errors Checker"
    deferred = False

    metric_name = "design__lint_error__count"
    metric_description = "Lint errors"


@Step.factory.register()
class LintWarnings(MetricChecker):
    id = "Checker.LintWarnings"
    name = "Lint Warnings Checker"
    long_name = "Lint Warnings Checker"
    deferred = False

    metric_name = "design__lint_warning__count"
    metric_description = "Lint warnings"


@Step.factory.register()
class LintTimingConstructs(MetricChecker):
    id = "Checker.LintTimingConstructs"
    name = "Lint Timing Error Checker"
    long_name = "Lint Timing Errors Checker"
    deferred = False

    metric_name = "design__lint_timing_construct__count"
    metric_description = "Lint Timing Errors"

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        metric_value = state_in.metrics.get(self.metric_name)

        if metric_value is not None:
            if metric_value > 0:
                error_msg = "Timing constructs found in the RTL. Please remove them or wrap them around an ifdef. It heavily unrecommended to rely on timing constructs for synthesis."
                err(f"{error_msg}")
                raise StepError(error_msg)
            else:
                info(f"Check for {self.metric_description} clear.")
        else:
            warn(
                f"The {self.metric_description} metric was not found. Are you sure the relevant step was run?"
            )

        return {}, {}


@Step.factory.register()
class KLayoutDRC(MetricChecker):
    id = "Checker.KLayoutDRC"
    name = "KLayout DRC Checker"
    long_name = "KLayout Design Rule Checker"

    metric_name = "klayout__drc_error__count"
    metric_description = "KLayout DRC errors"


class TimingViolations(MetricChecker):
    name = "Timing Violations Checker"
    long_name = "Timing Violations Checker"
    violation_type: str = NotImplemented

    config_vars = [
        Variable(
            "TIMING_VIOLATIONS_CORNERS",
            List[str],
            "A list of wildcards matching IPVT corners to use during checking for timing violations.",
            default=["*tt*"],
        ),
    ]

    def check_timing_violations(
        self,
        metric_basename: str,
        state_in: State,
        threshold: Optional[Decimal],
        violation_type: str,
    ):
        if not threshold:
            threshold = Decimal(0)

        metrics = {
            key: value
            for key, value in state_in.metrics.items()
            if metric_basename in key
        }
        debug("metrics ▶")
        debug(metrics)
        if not metrics:
            warn(f"No metrics found for {metric_basename}")
        else:
            metric_corners = set([key.split(":")[1] for key in metrics.keys()])
            unmatched_config_corners = set(
                [
                    config_corner
                    for config_corner in self.config["TIMING_VIOLATIONS_CORNERS"]
                    if not [
                        corner
                        for corner in metric_corners
                        if fnmatch.fnmatch(corner, config_corner)
                    ]
                ]
            )
            matched_corners = set(
                [
                    corner
                    for corner in metric_corners
                    if [
                        config_corner
                        for config_corner in self.config["TIMING_VIOLATIONS_CORNERS"]
                        if fnmatch.fnmatch(corner, config_corner)
                    ]
                ]
            )
            violating_corners = [
                corner
                for corner in matched_corners
                if metrics[f"{metric_basename}:{corner}"] > threshold
            ]

            debug("corners ▶")
            debug(metric_corners)
            debug("unmatched config corners ▶")
            debug(unmatched_config_corners)
            debug("matched corners ▶")
            debug(matched_corners)
            debug("violating corners ▶")
            debug(violating_corners)

            if unmatched_config_corners:
                msg = "The following specified TIMING_VIOLATIONS_CORNERS:\n"
                msg += "\n".join(unmatched_config_corners)
                raise DeferredStepError(msg)
            if violating_corners:
                msg = f"{violation_type} violations found in the following corners:\n"
                msg += "\n".join(violating_corners)
                raise DeferredStepError(msg)
            else:
                verbose(f"No {violation_type} violations found")

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        self.check_timing_violations(
            f"{self.metric_name}__corner",
            state_in,
            self.get_threshold(),
            self.violation_type,
        )

        return {}, {}


@Step.factory.register()
class SetupViolations(TimingViolations):
    id = "Checker.SetupViolations"
    name = "Setup Timing Violations Checker"
    long_name = "Setup Timing Violations Checker"
    violation_type = "setup"

    metric_name = "timing__setup_vio__count"


@Step.factory.register()
class HoldViolations(TimingViolations):
    id = "Checker.HoldViolations"
    name = "Hold Timing Violations Checker"
    long_name = "Hold Timing Violations Checker"
    violation_type = "hold"

    metric_name = "timing__hold_vio__count"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import subprocess
from typing import List, Optional, Tuple

from .step import Step, StepException, ViewsUpdate, MetricsUpdate
from ..config import Variable
from ..state import DesignFormat, State
from ..common import Path


@Step.factory.register()
class Lint(Step):
    """Lint design Verilog source files"""

    id = "Verilator.Lint"
    name = "Verilator Lint"
    long_name = "Verilator Lint"
    inputs = []  # The input RTL is part of the configuration
    outputs = []

    config_vars = [
        Variable(
            "VERILOG_FILES",
            List[Path],
            "The paths of the design's Verilog files.",
        ),
        Variable(
            "LINTER_INCLUDE_PDK_MODELS",
            bool,
            "Include Verilog models of the PDK",
            default=False,
        ),
        Variable(
            "LINTER_RELATIVE_INCLUDES",
            bool,
            "When a file references an include file, resolve the filename relative to the path of the referencing file, instead of relative to the current directory.",
            default=True,
            deprecated_names=["VERILATOR_RELATIVE_INCLUDES"],
        ),
        Variable(
            "LINTER_ERROR_ON_LATCH",
            bool,
            "When a latch is inferred by an `always` block that is not explicitly marked as `always_latch`, report this as a linter error.",
            default=True,
        ),
        Variable(
            "VERILOG_DEFINES",
            Optional[List[str]],
            "Preprocessor defines for input Verilog files",
            deprecated_names=["SYNTH_DEFINES"],
        ),
        Variable(
            "LINTER_DEFINES",
            Optional[List[str]],
            "Linter-specific preprocessor definitions; overrides VERILOG_DEFINES for the lint step if exists",
        ),
        Variable(
            "QUIT_ON_LINTER_WARNINGS",
            bool,
            "Quit on linter warnings.",
            default=False,
            deprecated_names=["QUIT_ON_VERILATOR_WARNINGS"],
        ),
    ]

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        kwargs, env = self.extract_env(kwargs)
        views_updates: ViewsUpdate = {}
        metrics_updates: MetricsUpdate = {}
        extra_args = []

        scl_models = (
            self.config["CELL_VERILOG_MODELS"]
            or self.config["CELL_BB_VERILOG_MODELS"]
            or []
        )
        input_models = (
            scl_models
            + self.toolbox.get_macro_views(self.config, DesignFormat.NETLIST)
            + (self.config["EXTRA_VERILOG_MODELS"] or [])
        )  # not +=: break reference!

        defines = self.config["LINTER_DEFINES"] or self.config["VERILOG_DEFINES"] or []

        bb_path = self.toolbox.create_blackbox_model(
            frozenset([str(path) for path in input_models]),
            frozenset(defines),
        )

        bb_with_guards = os.path.join(self.step_dir, "bb.v")
        with open(bb_path, "r", encoding="utf8",) as bb_in, open(
            bb_with_guards,
            "w",
            encoding="utf8",
        ) as bb_out:
            "\n/* verilator lint_off UNUSEDSIGNAL */\n$out_str\n/* verilator lint_on UNUSEDSIGNAL */\n/* verilator lint_on UNDRIVEN */"
            print("/* verilator lint_off UNDRIVEN */", file=bb_out)
            print("/* verilator lint_off UNUSEDSIGNAL */", file=bb_out)
            for line in bb_in:
                bb_out.write(line)
            print("/* verilator lint_on UNDRIVEN */", file=bb_out)
            print("/* verilator lint_on UNUSEDSIGNAL */", file=bb_out)

        if not self.config["QUIT_ON_LINTER_WARNINGS"]:
            extra_args.append("--Wno-fatal")

        if self.config["LINTER_RELATIVE_INCLUDES"]:
            extra_args.append("--relative-includes")

        if self.config["LINTER_ERROR_ON_LATCH"]:
            extra_args.append("--Werror-LATCH")

        for define in defines:
            extra_args.append(f"+define+{define}")

        exit_error: Optional[subprocess.CalledProcessError] = None
        try:
            self.run_subprocess(
                [
                    "verilator",
                    "--lint-only",
                    "--Wall",
                    "--Wno-DECLFILENAME",
                    "--Wno-EOFNEWLINE",
                    "--top-module",
                    self.config["DESIGN_NAME"],
                ]
                + [bb_with_guards]
                + self.config["VERILOG_FILES"]
                + extra_args,
                env=env,
            )
        except subprocess.CalledProcessError as e:
            exit_error = e

        warnings_count = 0
        errors_count = 0
        latch_count = 0
        timing_constructs = 0

        exiting_rx = re.compile(r"^\s*%Error: Exiting due to (\d+) error\(s\)")
        with open(self.get_log_path(), "r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if r"%Warning-" in line:
                    warnings_count += 1
                if r"%Error-LATCH" in line or r"%Warning-LATCH" in line:
                    latch_count += 1
                if r"%Error-NEEDTIMINGOPT" in line:
                    timing_constructs += 1
                if match := exiting_rx.search(line):
                    errors_count = int(match[1])

            if exit_error is not None and errors_count == 0:
                raise StepException(f"Verilator exited unexpectedly: {exit_error}")

        metrics_updates.update({"design__lint_error__count": errors_count})
        metrics_updates.update(
            {"design__lint_timing_construct__count": timing_constructs}
        )
        metrics_updates.update({"design__lint_warning__count": warnings_count})
        metrics_updates.update({"design__inferred_latch__count": latch_count})
        return views_updates, metrics_updates

    def layout_preview(self) -> Optional[str]:
        return None



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Type

from .flow import Flow
from .sequential import SequentialFlow
from ..state import DesignFormat
from ..config import Variable
from ..steps import (
    Step,
    Yosys,
    OpenROAD,
    Magic,
    KLayout,
    Odb,
    Netgen,
    Checker,
    Verilator,
    Misc,
)

_common_config_vars = [
    Variable(
        "RUN_TAP_ENDCAP_INSERTION",
        bool,
        "Enables the OpenROAD.TapEndcapInsertion step.",
        default=True,
        deprecated_names=["TAP_DECAP_INSERTION", "RUN_TAP_DECAP_INSERTION"],
    ),
    Variable(
        "RUN_POST_GPL_DESIGN_REPAIR",
        bool,
        "Enables resizer design repair after global placement using the OpenROAD.RepairDesignPostGPL step.",
        default=True,
        deprecated_names=["PL_RESIZER_DESIGN_OPTIMIZATIONS", "RUN_REPAIR_DESIGN"],
    ),
    Variable(
        "RUN_POST_GRT_DESIGN_REPAIR",
        bool,
        "Enables resizer design repair after global placement using the OpenROAD.RepairDesignPostGPL step. This is experimental and may result in hangs and/or extended run times.",
        default=False,
    ),
    Variable(
        "RUN_CTS",
        bool,
        "Enables clock tree synthesis using the OpenROAD.CTS step.",
        default=True,
        deprecated_names=["CLOCK_TREE_SYNTH"],
    ),
    Variable(
        "RUN_POST_CTS_RESIZER_TIMING",
        bool,
        "Enables resizer timing optimizations after clock tree synthesis using the OpenROAD.ResizerTimingPostCTS step.",
        default=True,
        deprecated_names=["PL_RESIZER_TIMING_OPTIMIZATIONS"],
    ),
    Variable(
        "RUN_POST_GRT_RESIZER_TIMING",
        bool,
        "Enables resizer timing optimizations after global routing using the OpenROAD.ResizerTimingPostGRT step. This is experimental and may result in hangs and/or extended run times.",
        default=False,
        deprecated_names=["GLB_RESIZER_TIMING_OPTIMIZATIONS"],
    ),
    Variable(
        "RUN_HEURISTIC_DIODE_INSERTION",
        bool,
        "Enables the Odb.HeuristicDiodeInsertion step.",
        default=False,  # For compatibility with OL1
    ),
    Variable(
        "RUN_ANTENNA_REPAIR",
        bool,
        "Enables the OpenROAD.RepairAntennas step.",
        default=True,
        deprecated_names=["GRT_REPAIR_ANTENNAS"],
    ),
    Variable(
        "RUN_DRT",
        bool,
        "Enables the OpenROAD.DetailedRouting step.",
        default=True,
    ),
    Variable(
        "RUN_FILL_INSERTION",
        bool,
        "Enables the OpenROAD.FillInsertion step.",
        default=True,
    ),
    Variable(
        "RUN_MCSTA",
        bool,
        "Enables multi-corner static timing analysis using the OpenROAD.STAPostPNR step.",
        default=True,
        deprecated_names=["RUN_SPEF_STA"],
    ),
    Variable(
        "RUN_SPEF_EXTRACTION",
        bool,
        "Enables parasitics extraction using the OpenROAD.RCX step.",
        default=True,
    ),
    Variable(
        "RUN_IRDROP_REPORT",
        bool,
        "Enables generation of an IR Drop report using the OpenROAD.IRDropReport step.",
        default=True,
    ),
    Variable(
        "RUN_LVS",
        bool,
        "Enables the Netgen.LVS step.",
        default=True,
    ),
    Variable(
        "RUN_MAGIC_STREAMOUT",
        bool,
        "Enables the Magic.StreamOut step to generate GDSII.",
        default=True,
        deprecated_names=["RUN_MAGIC"],
    ),
    Variable(
        "RUN_KLAYOUT_STREAMOUT",
        bool,
        "Enables the KLayout.StreamOut step to generate GDSII.",
        default=True,
        deprecated_names=["RUN_KLAYOUT"],
    ),
    Variable(
        "RUN_MAGIC_WRITE_LEF",
        bool,
        "Enables the Magic.WriteLEF step.",
        default=True,
        deprecated_names=["MAGIC_GENERATE_LEF"],
    ),
    Variable(
        "RUN_KLAYOUT_XOR",
        bool,
        "Enables running the KLayout.XOR step on the two GDSII files generated by Magic and Klayout. Stream-outs for both KLayout and Magic should have already run, and the PDK must support both signoff tools.",
        default=True,
    ),
    Variable(
        "RUN_MAGIC_DRC",
        bool,
        "Enables the Magic.DRC step.",
        default=True,
    ),
    Variable(
        "RUN_KLAYOUT_DRC",
        bool,
        "Enables the KLayout.DRC step.",
        default=True,
    ),
    Variable(
        "QUIT_ON_UNMAPPED_CELLS",
        bool,
        "Checks for unmapped cells after synthesis and quits immediately if so.",
        deprecated_names=["CHECK_UNMAPPED_CELLS"],
        default=True,
    ),
    Variable(
        "QUIT_ON_SYNTH_CHECKS",
        bool,
        "Quits the flow immediately if one or more synthesis check errors are flagged. This checks for combinational loops and/or wires with no drivers.",
        default=True,
    ),
    Variable(
        "QUIT_ON_TR_DRC",
        bool,
        "Checks for DRC violations after routing and exits the flow if any was found.",
        default=True,
    ),
    Variable(
        "QUIT_ON_MAGIC_DRC",
        bool,
        "Checks for DRC violations after magic DRC is executed and exits the flow if any was found.",
        default=True,
    ),
    Variable(
        "QUIT_ON_DISCONNECTED_PINS",
        bool,
        "Checks for disconnected instance pins after detailed routing and quits immediately if so.",
        default=True,
    ),
    Variable(
        "QUIT_ON_LONG_WIRE",
        bool,
        "Checks if any wire length exceeds the threshold set in the PDK. If so, an error is raised at the end of the flow.",
        default=False,
    ),
    Variable(
        "QUIT_ON_XOR_ERROR",
        bool,
        "Checks for geometric differences between the Magic and KLayout stream-outs. If any exist, raise an error at the end of the flow.",
        default=True,
    ),
    Variable(
        "QUIT_ON_ILLEGAL_OVERLAPS",
        bool,
        "Checks for illegal overlaps during Magic extraction. In some cases, these imply existing undetected shorts in the design. It raises an error at the end of the flow if so.",
        default=True,
    ),
    Variable(
        "QUIT_ON_LVS_ERROR",
        bool,
        "Checks for LVS errors after Netgen is executed. If any exist, it raises an error at the end of the flow.",
        default=True,
    ),
    Variable(
        "QUIT_ON_KLAYOUT_DRC",
        bool,
        "Checks for DRC violations after KLayout DRC is executed and exits the flow if any was found.",
        default=True,
    ),
    Variable(
        "QUIT_ON_PDN_VIOLATIONS",
        bool,
        "Checks for unconnected nodes in the power grid. If any exists, an error is raised at the end of the flow.",
        default=True,
        deprecated_names=["FP_PDN_CHECK_NODES"],
    ),
    Variable(
        "QUIT_ON_SETUP_VIOLATIONS",
        bool,
        "Check for setup violations in selected corners",
        default=True,
    ),
    Variable(
        "QUIT_ON_HOLD_VIOLATIONS",
        bool,
        "Check for setup violations in selected corners",
        default=True,
    ),
    Variable(
        "QUIT_ON_TIMING_VIOLATIONS",
        bool,
        "Check for timing violations in selected corners",
        default=True,
    ),
]

_common_gating_config_vars = {
    "OpenROAD.RepairDesignPostGPL": ["RUN_POST_GPL_DESIGN_REPAIR"],
    "OpenROAD.RepairDesignPostGRT": ["RUN_POST_GRT_DESIGN_REPAIR"],
    "OpenROAD.ResizerTimingPostCTS": ["RUN_POST_CTS_RESIZER_TIMING"],
    "OpenROAD.ResizerTimingPostGRT": ["RUN_POST_GRT_RESIZER_TIMING"],
    "OpenROAD.CTS": ["RUN_CTS"],
    "OpenROAD.RCX": ["RUN_SPEF_EXTRACTION"],
    "OpenROAD.TapEndcapInsertion": ["RUN_TAP_ENDCAP_INSERTION"],
    "Odb.HeuristicDiodeInsertion": ["RUN_HEURISTIC_DIODE_INSERTION"],
    "OpenROAD.RepairAntennas": ["RUN_ANTENNA_REPAIR"],
    "OpenROAD.DetailedRouting": ["RUN_DRT"],
    "OpenROAD.FillInsertion": ["RUN_FILL_INSERTION"],
    "OpenROAD.STAPostPNR": ["RUN_MCSTA"],
    "OpenROAD.IRDropReport": ["RUN_IRDROP_REPORT"],
    "Magic.StreamOut": ["RUN_MAGIC_STREAMOUT"],
    "KLayout.StreamOut": ["RUN_KLAYOUT_STREAMOUT"],
    "Magic.WriteLEF": ["RUN_MAGIC_WRITE_LEF"],
    "Magic.DRC": ["RUN_MAGIC_DRC"],
    "KLayout.DRC": ["RUN_KLAYOUT_DRC"],
    "KLayout.XOR": [
        "RUN_KLAYOUT_XOR",
        "RUN_MAGIC_STREAMOUT",
        "RUN_KLAYOUT_STREAMOUT",
    ],
    "Netgen.LVS": ["RUN_LVS"],
    "Checker.YosysUnmappedCells": ["QUIT_ON_UNMAPPED_CELLS"],
    "Checker.YosysSynthChecks": ["QUIT_ON_SYNTH_CHECKS"],
    "Checker.TrDRC": ["RUN_DRT", "QUIT_ON_TR_DRC"],
    "Checker.MagicDRC": ["RUN_MAGIC_DRC", "QUIT_ON_MAGIC_DRC"],
    "Checker.DisconnectedPins": ["QUIT_ON_DISCONNECTED_PINS"],
    "Checker.WireLength": ["QUIT_ON_LONG_WIRE"],
    "Checker.XOR": [
        "QUIT_ON_XOR_ERROR",
        "RUN_KLAYOUT_XOR",
        "RUN_MAGIC_STREAMOUT",
        "RUN_KLAYOUT_STREAMOUT",
    ],
    "Checker.IllegalOverlap": ["QUIT_ON_ILLEGAL_OVERLAPS"],
    "Checker.LVS": ["RUN_LVS", "QUIT_ON_LVS_ERROR"],
    "Checker.KLayoutDRC": ["RUN_KLAYOUT_DRC", "QUIT_ON_KLAYOUT_DRC"],
    "Checker.PowerGridViolations": ["QUIT_ON_PDN_VIOLATIONS"],
    "Checker.SetupViolations": [
        "QUIT_ON_TIMING_VIOLATIONS",
        "QUIT_ON_SETUP_VIOLATIONS",
    ],
    "Checker.HoldViolations": ["QUIT_ON_TIMING_VIOLATIONS", "QUIT_ON_HOLD_VIOLATIONS"],
}


@Flow.factory.register()
class Classic(SequentialFlow):
    """
    A flow of type :class:`openlane.flows.SequentialFlow` that is the most
    similar to the original OpenLane 1.0 flow, running the Verilog RTL through
    Yosys, OpenROAD, KLayout and Magic to produce a valid GDSII for simpler designs.

    This is the default when using OpenLane via the command-line.
    """

    Steps = [
        Verilator.Lint,
        Checker.LintTimingConstructs,
        Checker.LintErrors,
        Checker.LintWarnings,
        Yosys.JsonHeader,
        Yosys.Synthesis,
        Checker.YosysUnmappedCells,
        Checker.YosysSynthChecks,
        OpenROAD.CheckSDCFiles,
        OpenROAD.CheckMacroInstances,
        OpenROAD.STAPrePNR,
        OpenROAD.Floorplan,
        Odb.CheckMacroAntennaProperties,
        Odb.SetPowerConnections,
        Odb.ManualMacroPlacement,
        OpenROAD.CutRows,
        OpenROAD.TapEndcapInsertion,
        OpenROAD.GlobalPlacementSkipIO,
        OpenROAD.IOPlacement,
        Odb.CustomIOPlacement,
        Odb.ApplyDEFTemplate,
        OpenROAD.GlobalPlacement,
        Odb.AddPDNObstructions,
        OpenROAD.GeneratePDN,
        Odb.RemovePDNObstructions,
        Odb.WriteVerilogHeader,
        Checker.PowerGridViolations,
        OpenROAD.STAMidPNR,
        OpenROAD.RepairDesignPostGPL,
        OpenROAD.DetailedPlacement,
        OpenROAD.CTS,
        OpenROAD.STAMidPNR,
        OpenROAD.ResizerTimingPostCTS,
        OpenROAD.STAMidPNR,
        OpenROAD.GlobalRouting,
        OpenROAD.CheckAntennas,
        OpenROAD.RepairDesignPostGRT,
        Odb.DiodesOnPorts,
        Odb.HeuristicDiodeInsertion,
        OpenROAD.RepairAntennas,
        OpenROAD.ResizerTimingPostGRT,
        OpenROAD.STAMidPNR,
        OpenROAD.DetailedRouting,
        OpenROAD.CheckAntennas,
        Checker.TrDRC,
        Odb.ReportDisconnectedPins,
        Checker.DisconnectedPins,
        Odb.ReportWireLength,
        Checker.WireLength,
        OpenROAD.FillInsertion,
        OpenROAD.RCX,
        OpenROAD.STAPostPNR,
        OpenROAD.IRDropReport,
        Magic.StreamOut,
        KLayout.StreamOut,
        Magic.WriteLEF,
        Odb.CheckDesignAntennaProperties,
        KLayout.XOR,
        Checker.XOR,
        Magic.DRC,
        KLayout.DRC,
        Checker.MagicDRC,
        Checker.KLayoutDRC,
        Magic.SpiceExtraction,
        Checker.IllegalOverlap,
        Netgen.LVS,
        Checker.LVS,
        Yosys.EQY,
        Checker.SetupViolations,
        Checker.HoldViolations,
        Misc.ReportManufacturability,
    ]

    config_vars = _common_config_vars + [
        Variable(
            "RUN_EQY",
            bool,
            "Enables the Yosys.EQY step.",
            default=False,
        ),
        Variable(
            "RUN_LINTER",
            bool,
            "Enables the Verilator.Lint step and associated checker steps.",
            default=True,
            deprecated_names=["RUN_VERILATOR"],
        ),
        Variable(
            "QUIT_ON_LINTER_ERRORS",
            bool,
            "Quit immediately on any linter errors.",
            default=True,
            deprecated_names=["QUIT_ON_VERILATOR_ERRORS"],
        ),
        Variable(
            "QUIT_ON_LINTER_WARNINGS",
            bool,
            "Quit immediately on any linter warnings.",
            default=False,
            deprecated_names=["QUIT_ON_VERILATOR_WARNINGS"],
        ),
        Variable(
            "QUIT_ON_LINTER_TIMING_CONSTRUCTS",
            bool,
            "Quit immediately on any discovered timing constructs during linting.",
            default=True,
            deprecated_names=["QUIT_ON_LINTER_TIMING_CONSTRUCTS"],
        ),
    ]

    gating_config_vars = _common_gating_config_vars.copy()


Classic.gating_config_vars.update(
    {
        "Yosys.EQY": ["RUN_EQY"],
        "Verilator.Lint": ["RUN_LINTER"],
        "Checker.LintErrors": ["RUN_LINTER", "QUIT_ON_LINTER_ERRORS"],
        "Checker.LintWarnings": ["RUN_LINTER", "QUIT_ON_LINTER_WARNINGS"],
        "Checker.LintTimingConstructs": [
            "RUN_LINTER",
            "QUIT_ON_LINTER_TIMING_CONSTRUCTS",
        ],
    }
)


def _vhdlclassic_substitute_verilog_steps(
    steps_in: List[Type[Step]],
) -> List[Type[Step]]:
    result: List[Type[Step]] = [Yosys.VHDLSynthesis]
    for step in steps_in:
        # Ignore Verilog header-dependent steps and such
        if (
            step.id.startswith("Yosys.")
            or step.id.startswith("Checker.Lint")
            or step.id.startswith("Verilator.")
            or DesignFormat.JSON_HEADER in step.inputs
        ):
            continue
        result.append(step)
    return result


@Flow.factory.register()
class VHDLClassic(SequentialFlow):
    """
    A variant of :class:`openlane.flows.Classic` that uses VHDL files as an
    input instead of Verilog.
    """

    Steps: List[Type[Step]] = _vhdlclassic_substitute_verilog_steps(Classic.Steps)

    config_vars = _common_config_vars.copy()
    gating_config_vars = _common_gating_config_vars.copy()
